@@ -1,10 +1,12 @@
+import hashlib
 import re
 import socket
 import time
 import traceback
+from base64 import urlsafe_b64encode
 from functools import update_wrapper
 from os.path import expanduser
-from subprocess import PIPE, STDOUT, Popen
+from subprocess import PIPE, STDOUT, DEVNULL, Popen
 from time import sleep
 from typing import Optional, Any, List
 
@@ -107,6 +109,93 @@ def _date_to_str(dt: Optional[datetime.datetime]) -> Optional[str]:
     return used_dt.isoformat()
 
 
+def _has_gftp() -> bool:
+    from shutil import which
+    return which('gftp') is not None
+
+
+def _get_short_file_short_hash_id(filename: str) -> str:
+    with open(filename, 'r') as f:
+        return urlsafe_b64encode(hashlib.sha1(f.read().encode('utf-8')).digest()).decode('utf-8').replace('=', '')[0:10]
+
+
+def _get_str_short_hash_id(content: str) -> str:
+    return urlsafe_b64encode(hashlib.sha1(content.encode('utf-8')).digest()).decode('utf-8').replace('=', '')[0:10]
+
+
+def _sh_quote_arg(c: str) -> str:
+    if "'" not in c:
+        if ' ' not in c and '"' not in c and '$' not in c:
+            return c
+        else:
+            return f"'{c}'"
+    else:
+        c = c.replace("'", """'"'"'""")
+        return f"'{c}'"
+
+
+def _command_array_to_str(command: List[str]) -> str:
+    return ' '.join([_sh_quote_arg(c) for c in command])
+
+
+def _add_gpulab_ssh_config_hack(
+        ctx,
+        target_hostname: str, target_identity: str,
+        proxy_hostname: Optional[str], proxy_port: Optional[int],
+        proxy_username: Optional[str], proxy_identity: Optional[str],
+) -> str:
+    """
+    Adds a new ~/.ssh/config Host that can be used to force the SSH proxy easily.
+    :param target_hostname:
+    :param proxy_hostname:
+    :return: the fake hostname
+    """
+    # we need to be sure the correct identity is used
+    target_identity_id = _get_short_file_short_hash_id(target_identity)
+    proxy_identity_id = _get_short_file_short_hash_id(proxy_identity) if proxy_identity else None
+
+    if proxy_hostname:
+        assert proxy_port
+        assert proxy_username
+        assert proxy_identity
+        id = f'{proxy_username}.{proxy_identity_id}.{proxy_hostname}.{target_identity_id}'
+        fake_hostname = f'gpulabproxyhack.{_get_str_short_hash_id(id)}.{target_hostname}'
+    else:
+        fake_hostname = f'gpulabnoproxyhack.{target_identity_id}.{target_hostname}'
+
+    from os.path import expanduser
+    ssh_dir = expanduser('~/.ssh/')
+    if not os.path.exists(ssh_dir):
+        os.mkdir(ssh_dir, 0o700)
+
+    ssh_config_filename = expanduser('~/.ssh/config')
+
+    # Check if Host already added
+    if os.path.exists(ssh_config_filename):
+        with open(ssh_config_filename, 'r') as f:
+            for line in f:
+                if line.startswith(f'Host={fake_hostname}'):
+                    return fake_hostname
+
+    created = os.path.exists(ssh_config_filename)
+    with open(ssh_config_filename, 'a') as f:
+        f.write(f'\n')
+        f.write(f'# This entry is added by GPULab. Feel free to delete it. GPULab will rewrite it if needed.\n')
+        f.write(f'Host={fake_hostname}\n')
+        f.write(f'   Hostname {target_hostname}\n')
+        if proxy_hostname:
+            f.write(f'   ProxyCommand ssh -i "{proxy_identity}" -oPort={proxy_port} "{proxy_username}@{proxy_hostname}" -W %h:%p\n')
+        f.write(f'   IdentityFile "{target_identity}"\n')
+        f.write(f'   IdentitiesOnly yes\n')
+        f.write(f'   PreferredAuthentications publickey\n')
+        f.write(f'\n')
+    if created:
+        os.chmod(ssh_config_filename, 0o700)
+    if ctx.obj[VERBOSE_DEBUG]:
+        click.secho(f'Appended Host={fake_hostname} to {ssh_config_filename}')
+    return fake_hostname
+
+
 def _print_table(headers: List[str], rows: List[List[Any]], add_vert_lines: bool = False, add_hor_lines: bool = False,
                  seperator : Optional[str] = None):
     rows_and_headers = list(rows)
@@ -181,7 +270,7 @@ def _port_mapping_format(port_mappings: List[Job2PortMapping]) -> str:
 
 CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
 
-GPULAB_VERSION = '3.1.0'
+GPULAB_VERSION = '3.2.0'
 
 
 def check_and_warn_version(ctx):
@@ -240,8 +329,9 @@ def check_and_warn_version(ctx):
         latest_is_older = False
 
     if not latest_is_older and latest_version != GPULAB_VERSION:
-        click.secho("A new version ({}) of gpulab-cli is available. (This is version {})"
-                    .format(latest_version, GPULAB_VERSION),
+        click.secho(f"A new version ({latest_version}) of gpulab-cli is available. (This is version {GPULAB_VERSION})",
+                    err=True, nl=True, fg='yellow')
+        click.secho("Upgrade with:  pip3 install --upgrade imecilabt-gpulab-cli",
                     err=True, nl=True, fg='yellow')
 
 
@@ -338,7 +428,7 @@ def modify_usage_error():
 @click.version_option(GPULAB_VERSION, message='%(version)s')
 @click.pass_context
 def cli(ctx: click.Context, cert, password, deployment_environment: str, servercert, debug, custom_master_url):
-    """GPULab client version 3.1.0
+    """GPULab client version 3.2.0
 
     \b
     This is the general help. For help with a specific command, try:
@@ -571,7 +661,7 @@ def print_bugreport_context(ctx):
 @click.version_option(GPULAB_VERSION, message='%(version)s')
 @click.pass_context
 def localhost_mode_cli(ctx):
-    """GPULab client version 3.1.0 (in LOCALHOST_MODE!)
+    """GPULab client version 3.2.0 (in LOCALHOST_MODE!)
 
     \b
     Send bugreports, questions and feedback to: gpulab@ilabt.imec.be
@@ -1534,7 +1624,7 @@ def actual_wait(ctx, job_id, wait_run, wait_done):
 
 @click.command("ssh", short_help="Log in to a Job's container using SSH.")
 @click.argument('job_id', required=True, type=click.STRING)
-@click.option('-s', '--show', is_flag=True, help="Do not login, just show the SSH command.")
+@click.option('-s', '--show', is_flag=True, help="Do not connect, only show the SSH command.")
 @click.option('-p', '--proxy', 'proxy', flag_value='yes', default=False,
               help="Use the jFed proxy to reach the container")
 @click.option('-P', '--no-proxy', 'proxy', flag_value='no', default=False,
@@ -1550,11 +1640,39 @@ def ssh(ctx, job_id, show, show_ansible, proxy):
     actual_ssh(ctx, job_id, show, show_ansible, proxy)
 
 
-def actual_ssh(ctx, job_id, show, show_ansible, proxy):
+@click.command("sftp", short_help="Access files in a Job's container using SFTP.")
+@click.argument('job_id', required=True, type=click.STRING)
+@click.option('-s', '--show', is_flag=True, help="Do not connect, only show the SFTP command.")
+@click.option('-p', '--proxy', 'proxy', flag_value='yes', default=False,
+              help="Use the jFed proxy to reach the container")
+@click.option('-P', '--no-proxy', 'proxy', flag_value='no', default=False,
+              help="Don't use the jFed proxy to reach the container (requires IPv6 or VPN)")
+@click.option('--auto-proxy', 'proxy', flag_value='auto', default=True,
+              help="Automatically use the jFed proxy if needed to reach the container (Default)")
+@click.option('-f', '--sftp', 'sftp_command', flag_value='sftp', default=False,
+              help="Use sftp command: sftp (CLI)")
+@click.option('-g', '--gftp', 'sftp_command', flag_value='gftp', default=False,
+              help="Use sftp command: gftp (GUI)")
+@click.option('--auto-sftp-command', 'sftp_command', flag_value='auto', default=True,
+              help="Use sftp command: automatic (Default)")
+@click.pass_context
+@require_ctx_loggedin_api_client
+def sftp(ctx, job_id, show, proxy, sftp_command):
+    if sftp_command == 'auto' and _has_gftp():
+        sftp_command = 'gftp'
+    else:
+        sftp_command = 'sftp'
+    actual_ssh(ctx, job_id, show, False, proxy,
+               sftp_command=sftp_command)
+
+
+def actual_ssh(ctx, job_id, show, show_ansible, proxy, *, sftp_command=None):
     """Log in to a Job's container using SSH."""
     client = ctx.obj[API_CLIENT]
     s = client.get_session()  # type: requests.Session
 
+    assert sftp_command is None or sftp_command in ('sftp', 'gftp')
+
     r = s.get(client.get_api_url(3) + "jobs/" + job_id, timeout=GPULAB_CLIENT_TIMEOUT)
 
     if r.ok:
@@ -1571,32 +1689,20 @@ def actual_ssh(ctx, job_id, show, show_ansible, proxy):
 
         if proxy == 'auto':
             # Try to resolve worker, if it doesn't assume we need the proxy
+            # (reason: IDLab Gent hostnames have public IPv6, but not IPv4.)
             try:
                 proxy_ip = socket.gethostbyname(job.state.resources.slave_host)
                 proxy = 'no'
+
+                # Also hardcode that uantwerpen needs proxy.
+                # (reason: uantwerpen nodes are firewalled)
+                if job.state.resources.slave_host.endswith('.idlab.uantwerpen.be'):
+                    proxy = 'yes'
             # except socket.gaierror as e:
             except OSError as e:
                 proxy = 'yes'
 
-        proxy_argument = ''
-        if proxy == 'yes':
-            proxy_username = job.state.resources.ssh_proxy_username
-            proxy_host = job.state.resources.ssh_proxy_host
-            proxy_port = job.state.resources.ssh_proxy_port
-            if not proxy_username:
-                raise click.ClickException("--proxy option cannot be used: no SSH proxy info in job. "
-                                           "(You could try again with --no-proxy but that requires IPv6 or VPN.)")
-            proxy_argument = ' -oProxyCommand="ssh -i \'{}\'{} {}@{} -W %h:%p"' \
-                .format(ctx.obj[CERTFILE], _ssh_port_command(proxy_port), proxy_username, proxy_host)
-
-        if show:
-            ssh_command = 'ssh -i \'{}\'{} {}@{}{}'.format(
-                ctx.obj[CERTFILE],
-                _ssh_port_command(job.state.resources.ssh_port),
-                job.state.resources.ssh_username, job.state.resources.ssh_host,
-                proxy_argument)
-            click.echo(ssh_command)
-        elif show_ansible:
+        if show_ansible:
             # Ansible can be used with the proxy, but it requires more than a line in the inventory file:
             #   it requires adding an ssh-config file in ansible.cfg and adding the ProxyCommand for the ansible
             #   node in the ssh-config file.
@@ -1604,11 +1710,75 @@ def actual_ssh(ctx, job_id, show, show_ansible, proxy):
                 .format(ctx.obj[CERTFILE], job.state.resources.slave_host, job.state.resources.ssh_username)
             click.echo(ssh_command)
         else:
-            command = ['ssh', '-i', ctx.obj[CERTFILE], '-oPort={}'.format(job.state.resources.ssh_port)]
-            if proxy_argument.replace('"', '').strip():
-                command.append(proxy_argument.replace('"', '').strip())
-            command.append('{}@{}'.format(job.state.resources.ssh_username, job.state.resources.ssh_host))
-            os.execvp('ssh', command)
+            if sftp_command == 'gftp':
+                if proxy == 'yes':
+                    fake_hostname = _add_gpulab_ssh_config_hack(
+                        ctx,
+                        target_hostname=job.state.resources.ssh_host,
+                        target_identity=ctx.obj[CERTFILE],
+                        proxy_hostname=job.state.resources.ssh_proxy_host,
+                        proxy_port=job.state.resources.ssh_proxy_port,
+                        proxy_username=job.state.resources.ssh_proxy_username,
+                        proxy_identity=ctx.obj[CERTFILE],
+                    )
+                else:
+                    fake_hostname = _add_gpulab_ssh_config_hack(
+                        ctx,
+                        target_hostname=job.state.resources.ssh_host,
+                        target_identity=ctx.obj[CERTFILE],
+                        proxy_hostname=None,
+                        proxy_port=None,
+                        proxy_username=None,
+                        proxy_identity=None,
+                    )
+                if ctx.obj[VERBOSE_DEBUG]:
+                    click.secho(f'Configured fake hostname for use with gftp: '
+                                f'fake={fake_hostname!r} -> real={job.state.resources.ssh_host}')
+
+                command = [
+                    'gftp',
+                    'ssh2://{}@{}:{}'.format(
+                        job.state.resources.ssh_username,
+                        fake_hostname,  # instead of: job.state.resources.ssh_host,
+                        job.state.resources.ssh_port)
+                ]
+
+
+                # with Popen(command, stderr=DEVNULL, stdout=DEVNULL, universal_newlines=True) as proc:
+                #     # for l in proc.stdout:
+                #     #     click.secho(l)
+                #     res_code = proc.wait()
+
+                if show:
+                    click.echo(_command_array_to_str(command))
+                else:
+                    if ctx.obj[VERBOSE_DEBUG]:
+                        click.secho(f'Calling: {command!r}')
+                    # Note: This REPLACES the current python process!
+                    os.execvp('gftp', command)
+            else:
+                c = 'ssh' if sftp_command is None else 'sftp'
+                command = [
+                    c,
+                    '-i', ctx.obj[CERTFILE],
+                    '-oPort={}'.format(job.state.resources.ssh_port)]
+
+                if proxy == 'yes':
+                    proxy_username = job.state.resources.ssh_proxy_username
+                    proxy_host = job.state.resources.ssh_proxy_host
+                    proxy_port = job.state.resources.ssh_proxy_port
+                    if not proxy_username:
+                        raise click.ClickException("--proxy option cannot be used: no SSH proxy info in job. "
+                                                   "(You could try again with --no-proxy but that requires IPv6 or VPN.)")
+                    command.append('-oProxyCommand=ssh -i {} {} {}@{} -W %h:%p'
+                                   .format(_sh_quote_arg(ctx.obj[CERTFILE]), _ssh_port_command(proxy_port), proxy_username,
+                                           proxy_host))
+                command.append('{}@{}'.format(job.state.resources.ssh_username, job.state.resources.ssh_host))
+                if show:
+                    click.echo(_command_array_to_str(command))
+                else:
+                    # Note: This REPLACES the current python process!
+                    os.execvp(c, command)
     else:
         raise click.ClickException("Could not retrieve info for job '{}': {}".format(job_id, r.text))
 
@@ -1659,6 +1829,7 @@ cli.add_command(get_slave_info)
 cli.add_command(get_info)
 cli.add_command(wait)
 cli.add_command(ssh)
+cli.add_command(sftp)
 cli.add_command(run_interactive)
 cli.add_command(print_bugreport_context)
 cli.add_command(convert_to_job2)
@@ -1675,6 +1846,7 @@ localhost_mode_cli.add_command(get_slave_info)
 localhost_mode_cli.add_command(get_info)
 localhost_mode_cli.add_command(wait)
 localhost_mode_cli.add_command(ssh)
+localhost_mode_cli.add_command(sftp)
 localhost_mode_cli.add_command(run_interactive)
 localhost_mode_cli.add_command(print_bugreport_context)
 localhost_mode_cli.add_command(convert_to_job2)
